import 'dart:ui';

import 'package:flutter/cupertino.dart';
import 'package:flutter/foundation.dart';
import 'package:flutter/foundation.dart' show kIsWeb;
import 'package:flutter/gestures.dart';
import 'package:flutter/material.dart';
import 'package:flutter/services.dart';
import 'package:flutter/widgets.dart';
import 'package:forme/forme.dart';

import '../../../forme_fields.dart';

typedef FormeAsyncChipInputBuilder<T> = Widget Function(
    T option, VoidCallback? onDelete);

typedef FormeAsyncChipInputFieldViewBuilder<T> = Widget Function(
  List<T> selected,
  Widget asyncAutocomplete,
  void Function(T option)? delete,
);

class FormeAsyncChipInput<T extends Object> extends FormeField<List<T>> {
  final FormeAsyncChipInputBuilder<T>? inputChipBuilder;
  final AutocompleteOptionToString<T> displayStringForOption;
  FormeAsyncChipInput({
    List<T> initialValue = const [],
    Key? key,
    required String name,
    bool readOnly = false,
    int? order,
    bool quietlyValidate = false,
    Duration? asyncValidatorDebounce,
    AutovalidateMode? autovalidateMode,
    FormeValueChanged<List<T>>? onValueChanged,
    FormeFocusChanged<List<T>>? onFocusChanged,
    FormeFieldSetter<List<T>>? onSaved,
    FormeValidator<List<T>>? validator,
    FormeAsyncValidator<List<T>>? asyncValidator,
    FormeFieldValidationChanged<List<T>>? onValidationChanged,
    required FormeAsyncAutocompleteOptionsBuilder<T> optionsBuilder,
    FormeAutocompleteOptionsViewBuilder<T>? optionsViewBuilder,
    double optionsMaxHeight = 200,
    InputDecoration? decoration,
    TextInputType? keyboardType,
    bool autofocus = false,
    int? maxLength,
    TextStyle? style,
    ToolbarOptions? toolbarOptions,
    TextInputAction? textInputAction,
    TextCapitalization textCapitalization = TextCapitalization.none,
    bool obscureText = false,
    StrutStyle? strutStyle,
    TextAlign textAlign = TextAlign.start,
    TextAlignVertical? textAlignVertical,
    TextDirection? textDirection,
    bool? showCursor,
    String obscuringCharacter = '•',
    bool autocorrect = true,
    SmartDashesType? smartDashesType,
    SmartQuotesType? smartQuotesType,
    bool enableSuggestions = true,
    bool expands = false,
    MaxLengthEnforcement? maxLengthEnforcement,
    double cursorWidth = 2.0,
    double? cursorHeight,
    Radius? cursorRadius,
    Color? cursorColor,
    BoxHeightStyle selectionHeightStyle = BoxHeightStyle.tight,
    BoxWidthStyle selectionWidthStyle = BoxWidthStyle.tight,
    Brightness? keyboardAppearance,
    EdgeInsets scrollPadding = const EdgeInsets.all(20),
    DragStartBehavior dragStartBehavior = DragStartBehavior.start,
    MouseCursor? mouseCursor,
    ScrollPhysics? scrollPhysics,
    Iterable<String>? autofillHints,
    bool enableInteractiveSelection = true,
    bool enabled = true,
    VoidCallback? onEditingComplete,
    List<TextInputFormatter>? inputFormatters,
    AppPrivateCommandCallback? appPrivateCommandCallback,
    InputCounterWidgetBuilder? buildCounter,
    GestureTapCallback? onTap,
    ScrollController? scrollController,
    TextSelectionControls? textSelectionControls,
    bool enableIMEPersonalizedLearning = true,
    FormeSearchCondition? searchCondition,
    FormeFieldDecorator<List<T>>? decorator,
    bool registrable = true,
    this.inputChipBuilder,
    this.displayStringForOption = RawAutocomplete.defaultStringForOption,
    InputDecoration? searchInputDecoration,
    Axis wrapDirection = Axis.horizontal,
    WrapAlignment wrapAlignment = WrapAlignment.start,
    double wrapSpacing = 4.0,
    WrapAlignment wrapRunAlignment = WrapAlignment.start,
    double wrapRunSpacing = 0.0,
    WrapCrossAlignment wrapCrossAxisAlignment = WrapCrossAlignment.center,
    TextDirection? wrapTextDirection,
    VerticalDirection wrapVerticalDirection = VerticalDirection.down,
    Clip wrapClipBehavior = Clip.none,
    double? searchInputStepWidth,
    double searchInputMinWidth = 50,
    FormeAsyncChipInputFieldViewBuilder<T>? fieldViewBuilder,
  }) : super(
            autovalidateMode: autovalidateMode,
            order: order,
            quietlyValidate: quietlyValidate,
            readOnly: readOnly,
            registrable: registrable,
            decorator: decorator ?? _DefaultDecorator(),
            key: key,
            asyncValidator: asyncValidator,
            asyncValidatorDebounce: asyncValidatorDebounce,
            onValidationChanged: onValidationChanged,
            onValueChanged: onValueChanged,
            onFocusChanged: onFocusChanged,
            onSaved: onSaved,
            validator: validator,
            name: name,
            initialValue: initialValue,
            builder: (genericState) {
              final _FormeAsyncChipInputState<T> state =
                  genericState as _FormeAsyncChipInputState<T>;
              void addValue(T value) {
                final List<T> values = List.of(state.value);
                if (!values.contains(value)) {
                  values.add(value);
                }
                state.didChange(values);
              }

              return FormeAsyncAutocomplete<T>(
                  optionsViewBuilder: optionsViewBuilder,
                  optionsMaxHeight: optionsMaxHeight,
                  showCursor: showCursor,
                  autofocus: autofocus,
                  autocorrect: autocorrect,
                  autofillHints: autofillHints,
                  buildCounter: buildCounter,
                  cursorColor: cursorColor,
                  cursorHeight: cursorHeight,
                  cursorRadius: cursorRadius,
                  cursorWidth: cursorWidth,
                  dragStartBehavior: dragStartBehavior,
                  enableIMEPersonalizedLearning: enableIMEPersonalizedLearning,
                  enableInteractiveSelection: enableInteractiveSelection,
                  enableSuggestions: enableSuggestions,
                  enabled: enabled,
                  expands: expands,
                  inputFormatters: inputFormatters,
                  keyboardAppearance: keyboardAppearance,
                  keyboardType: keyboardType,
                  maxLength: maxLength,
                  maxLengthEnforcement: maxLengthEnforcement,
                  mouseCursor: mouseCursor,
                  obscureText: obscureText,
                  obscuringCharacter: obscuringCharacter,
                  onEditingComplete: onEditingComplete,
                  onTap: onTap,
                  readOnly: state.readOnly,
                  scrollController: scrollController,
                  scrollPadding: scrollPadding,
                  scrollPhysics: scrollPhysics,
                  selectionHeightStyle: selectionHeightStyle,
                  selectionWidthStyle: selectionWidthStyle,
                  smartDashesType: smartDashesType,
                  smartQuotesType: smartQuotesType,
                  strutStyle: strutStyle,
                  style: style,
                  textAlign: textAlign,
                  textAlignVertical: textAlignVertical,
                  textCapitalization: textCapitalization,
                  textDirection: textDirection,
                  textInputAction: textInputAction,
                  textSelectionControls: textSelectionControls,
                  toolbarOptions: toolbarOptions,
                  decoration: searchInputDecoration,
                  displayStringForOption: displayStringForOption,
                  searchCondition: (TextEditingValue value) {
                    if (value.text.isEmpty) {
                      return false;
                    }

                    return searchCondition?.call(value) ?? true;
                  },
                  appPrivateCommandCallback: appPrivateCommandCallback,
                  registrable: false,
                  name: name,
                  onValueChanged: (f, v) {
                    if (v != null) {
                      addValue(v);
                      (f as FormeAsnycAutocompleteController)
                          .textFieldController
                          .text = '';
                    }
                  },
                  onFieldViewInitialed: (f) {
                    state.initFormeAsyncAutocomplete(
                        f as FormeAsnycAutocompleteController);
                  },
                  decorator: FormeInputDecoratorBuilder(
                      decoration: decoration?.copyWith(
                          suffixIcon: decoration.suffixIcon ??
                              Builder(
                                builder: (context) {
                                  final FormeAsnycAutocompleteController
                                      controller = FormeField.of<T?>(context)!
                                          as FormeAsnycAutocompleteController;
                                  return ValueListenableBuilder<
                                          FormeAsyncAutocompleteSearchState>(
                                      valueListenable:
                                          controller.stateListenable,
                                      builder: (context, state, child) {
                                        switch (state) {
                                          case FormeAsyncAutocompleteSearchState
                                              .loading:
                                            return const Align(
                                              alignment: Alignment.topCenter,
                                              widthFactor: 1,
                                              heightFactor: 1,
                                              child: SizedBox(
                                                  width: 18,
                                                  height: 18,
                                                  child:
                                                      CircularProgressIndicator(
                                                    strokeWidth: 3,
                                                  )),
                                            );
                                          case FormeAsyncAutocompleteSearchState
                                              .error:
                                            return const Align(
                                              alignment: Alignment.topCenter,
                                              widthFactor: 1,
                                              heightFactor: 1,
                                              child: Icon(Icons.dangerous,
                                                  color: Colors.redAccent),
                                            );
                                          case FormeAsyncAutocompleteSearchState
                                              .success:
                                          case FormeAsyncAutocompleteSearchState
                                              .waiting:
                                            return const SizedBox();
                                        }
                                      });
                                },
                              ),
                          suffixIconConstraints:
                              decoration.suffixIconConstraints ??
                                  const BoxConstraints.tightFor()),
                      emptyChecker: (value, controller) {
                        return state.value.isEmpty &&
                            (controller as FormeAsnycAutocompleteController)
                                .textFieldController
                                .text
                                .isEmpty;
                      },
                      wrapper: (child, controller) {
                        if (fieldViewBuilder != null) {
                          return fieldViewBuilder(
                              List.of(state.value),
                              child,
                              readOnly
                                  ? null
                                  : (T option) {
                                      if (state.value.contains(option)) {
                                        state.delete(option);
                                      }
                                    });
                        }
                        return Wrap(
                            direction: wrapDirection,
                            alignment: wrapAlignment,
                            runAlignment: wrapRunAlignment,
                            runSpacing: wrapRunSpacing,
                            spacing: wrapSpacing,
                            crossAxisAlignment: wrapCrossAxisAlignment,
                            textDirection: wrapTextDirection,
                            verticalDirection: wrapVerticalDirection,
                            clipBehavior: wrapClipBehavior,
                            children: state.value.map<Widget>((e) {
                              void onDeleted() {
                                state.delete(e);
                              }

                              if (inputChipBuilder != null) {
                                return inputChipBuilder(
                                    e, state.readOnly ? null : onDeleted);
                              }

                              return InputChip(
                                label: Text(displayStringForOption(e)),
                                onDeleted: state.readOnly ? null : onDeleted,
                              );
                            }).toList()
                              ..add(
                                IntrinsicWidth(
                                  stepWidth: searchInputStepWidth,
                                  child: Container(
                                    constraints: BoxConstraints(
                                        minWidth: searchInputMinWidth),
                                    child: child,
                                  ),
                                ),
                              ));
                      }),
                  optionsBuilder: (value) {
                    return optionsBuilder(value).then((value) {
                      return value
                          .where((element) => !state.value.contains(element));
                    });
                  });
            });

  @override
  FormeFieldState<List<T>> createState() => _FormeAsyncChipInputState();
}

class _FormeAsyncChipInputState<T extends Object>
    extends FormeFieldState<List<T>> {
  FormeAsnycAutocompleteController? asnycAutocompleteController;
  late final ValueNotifier<FormeAsyncAutocompleteSearchState> stateNotifier =
      FormeMountedValueNotifier(
          FormeAsyncAutocompleteSearchState.waiting, this);

  void initFormeAsyncAutocomplete(FormeAsnycAutocompleteController controller) {
    super.focusNode = controller.focusNode!;
    asnycAutocompleteController?.stateListenable.removeListener(onStateChange);
    asnycAutocompleteController = controller;
    controller.stateListenable.addListener(onStateChange);
    stateNotifier.value = controller.stateListenable.value;
  }

  @override
  FormeAsyncChipInput<T> get widget => super.widget as FormeAsyncChipInput<T>;

  void delete(T option) {
    final List<T> values = List.of(value);
    if (values.remove(option)) {
      didChange(values);
    }
  }

  @override
  void dispose() {
    asnycAutocompleteController?.stateListenable.removeListener(onStateChange);
    stateNotifier.dispose();
    super.dispose();
  }

  @override
  void reset() {
    super.reset();
    asnycAutocompleteController?.textFieldController.text = '';
  }

  void onStateChange() {
    if (asnycAutocompleteController != null) {
      stateNotifier.value = asnycAutocompleteController!.stateListenable.value;
    }
  }

  @override
  FormeFieldController<List<T>> createFormeFieldController() {
    return FormeAsyncChipInputController._(
        super.createFormeFieldController(), this);
  }
}

class FormeAsyncChipInputController<T> extends FormeFieldControllerDelegate<T> {
  final ValueListenable<FormeAsyncAutocompleteSearchState> stateListenable;
  FormeAsyncChipInputController._(
      FormeFieldController<T> delegate, _FormeAsyncChipInputState state)
      : stateListenable = _ValueListenableDelegate(state.stateNotifier),
        super(delegate);
}

class _DefaultDecorator<T extends Object> with FormeFieldDecorator<List<T>> {
  @override
  Widget build(FormeFieldController<List<T>> controller, Widget child) {
    return GestureDetector(
      child: child,
      onTap: () {
        controller.focusNode?.requestFocus();
        if (!controller.readOnly && !kIsWeb) {
          SystemChannels.textInput.invokeMethod<Object>('TextInput.show');
        }
      },
    );
  }
}

class _ValueListenableDelegate<T> extends ValueListenable<T> {
  final ValueNotifier<T> delegate;

  const _ValueListenableDelegate(this.delegate);

  @override
  void addListener(VoidCallback listener) => delegate.addListener(listener);

  @override
  void removeListener(VoidCallback listener) =>
      delegate.removeListener(listener);

  @override
  T get value => delegate.value;
}
